﻿using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.ComponentModel;
using System;
using System.Security.Cryptography.X509Certificates;
using System.IO;

namespace SignToolNet
{
    public class Program
    {
        // #define CRYPTUI_WIZ_NO_UI     1
        public const int CRYPTUI_WIZ_NO_UI = 1;

        // #define CRYPTUI_WIZ_DIGITAL_SIGN_SUBJECT_FILE     0x01
        public const int CRYPTUI_WIZ_DIGITAL_SIGN_SUBJECT_FILE = 1;

        // #define CRYPTUI_WIZ_DIGITAL_SIGN_CERT                    0x01
        public const int CRYPTUI_WIZ_DIGITAL_SIGN_CERT = 1;

        // #define CRYPTUI_WIZ_DIGITAL_SIGN_COMMERCIAL  0x0001
        public const int CRYPTUI_WIZ_DIGITAL_SIGN_COMMERCIAL = 1;

        // typedef struct _CRYPTUI_WIZ_DIGITAL_SIGN_INFO {  
        //   DWORD dwSize;  
        //   DWORD dwSubjectChoice;  
        //   union {    
        //       LPCWSTR pwszFileName;    
        //       PCCRYPTUI_WIZ_DIGITAL_SIGN_BLOB_INFO pSignBlobInfo;  
        //   };  
        //   DWORD dwSigningCertChoice;  
        //   union {    
        //       PCCERT_CONTEXT pSigningCertContext;    
        //       PCCRYPTUI_WIZ_DIGITAL_SIGN_STORE_INFO pSigningCertStore;    
        //       PCCRYPTUI_WIZ_DIGITAL_SIGN_CERT_PVK_INFO pSigningCertPvkInfo;  
        //   };  
        //   LPCWSTR pwszTimestampURL;  
        //   DWORD dwAdditionalCertChoice;  
        //   PCCRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO pSignExtInfo;
        // } CRYPTUI_WIZ_DIGITAL_SIGN_INFO;
        [StructLayout(LayoutKind.Sequential)]
        public struct CRYPTUI_WIZ_DIGITAL_SIGN_INFO
        {
            public int dwSize;
            public int dwSubjectChoice;
            [MarshalAs(UnmanagedType.LPWStr)]
            public string pwszFileName;
            public int dwSigningCertChoice;
            public IntPtr pSigningCertContext;
            [MarshalAs(UnmanagedType.LPWStr)]
            public string pwszTimestampURL;
            public int dwAdditionalCertChoice;
            public IntPtr pSignExtInfo;
        }

        //typedef struct _CRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO {
        //  DWORD             dwSize;
        //  DWORD             dwAttrFlags;
        //  LPCWSTR           pwszDescription;
        //  LPCWSTR           pwszMoreInfoLocation;
        //  LPCSTR            pszHashAlg;
        //  LPCWSTR           pwszSigningCertDisplayString;
        //  HCERTSTORE        hAdditionalCertStore;
        //  PCRYPT_ATTRIBUTES psAuthenticated;
        //  PCRYPT_ATTRIBUTES psUnauthenticated;
        //} CRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO, *PCRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO;
        [StructLayout(LayoutKind.Sequential)]
        public struct CRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO
        {
            public int dwSize;
            public int dwAttrFlags;
            [MarshalAs(UnmanagedType.LPWStr)]
            public string pwszDescription;
            [MarshalAs(UnmanagedType.LPWStr)]
            public string pwszMoreInfoLocation;
            [MarshalAs(UnmanagedType.LPStr)]
            public string pszHashAlg;
            [MarshalAs(UnmanagedType.LPWStr)]
            public string pwszSigningCertDisplayString;
            public IntPtr hAdditionalCertStore;
            public IntPtr psAuthenticated;
            public IntPtr psUnauthenticated;
        }

        // typedef struct _CRYPTUI_WIZ_DIGITAL_SIGN_CONTEXT {  
        //      DWORD dwSize;  
        //      DWORD cbBlob;  
        //      BYTE* pbBlob;
        // } CRYPTUI_WIZ_DIGITAL_SIGN_CONTEXT;
        [StructLayout(LayoutKind.Sequential)]
        public struct CRYPTUI_WIZ_DIGITAL_SIGN_CONTEXT
        {
            public int dwSize;
            public int cbBlob;
            public IntPtr pbBlob;
        }

        // BOOL WINAPI CryptUIWizDigitalSign(
        //      DWORD dwFlags,
        //      HWND hwndParent,
        //      LPCWSTR pwszWizardTitle,
        //      PCCRYPTUI_WIZ_DIGITAL_SIGN_INFO pDigitalSignInfo,
        //      PCCRYPTUI_WIZ_DIGITAL_SIGN_CONTEXT* ppSignContext
        // );
        [DllImport("Cryptui.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        public static extern bool CryptUIWizDigitalSign(
            int dwFlags,
            IntPtr hwndParent,
            [MarshalAs(UnmanagedType.LPWStr)]
            string pwszWizardTitle,
            ref CRYPTUI_WIZ_DIGITAL_SIGN_INFO pDigitalSignInfo,
            ref IntPtr ppSignContext);

        // BOOL WINAPI CryptUIWizFreeDigitalSignContext(
        //   PCCRYPTUI_WIZ_DIGITAL_SIGN_CONTEXT pSignContext
        // );
        [DllImport("Cryptui.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern bool CryptUIWizFreeDigitalSignContext(IntPtr pSignContext);

        public static void Main(string[] args)
        {
            var cla = new CommandLineArguments(args);
            try
            {
                var cert = new X509Certificate2(cla.Certificate, cla.Password);
                var digitalSignInfo = new CRYPTUI_WIZ_DIGITAL_SIGN_INFO();
                digitalSignInfo.dwSize = Marshal.SizeOf(digitalSignInfo);
                digitalSignInfo.dwSubjectChoice = CRYPTUI_WIZ_DIGITAL_SIGN_SUBJECT_FILE;
                digitalSignInfo.pwszFileName = cla.FileToSign;
                digitalSignInfo.dwSigningCertChoice = CRYPTUI_WIZ_DIGITAL_SIGN_CERT;
                digitalSignInfo.pSigningCertContext = cert.Handle;
                digitalSignInfo.pwszTimestampURL = cla.TimestampUrl;
                digitalSignInfo.dwAdditionalCertChoice = 0;
                digitalSignInfo.pSignExtInfo = IntPtr.Zero;

                // This should be needed for the /d and /du options but I haven't got it working yet.
                //var digitalSignExtInfo = new CRYPTUI_WIZ_DIGITAL_SIGN_EXTENDED_INFO();
                //digitalSignExtInfo.dwSize = Marshal.SizeOf(digitalSignExtInfo);
                //digitalSignExtInfo.dwAttrFlags = CRYPTUI_WIZ_DIGITAL_SIGN_COMMERCIAL;
                //digitalSignExtInfo.pwszDescription = cla.Description;
                //digitalSignExtInfo.pwszMoreInfoLocation = cla.MoreInfoUrl;
                //digitalSignExtInfo.pszHashAlg = null;
                //digitalSignExtInfo.pwszSigningCertDisplayString = string.Empty;
                //digitalSignExtInfo.hAdditionalCertStore = IntPtr.Zero;
                //digitalSignExtInfo.psAuthenticated = IntPtr.Zero;
                //digitalSignExtInfo.psUnauthenticated = IntPtr.Zero;

                //var pDigitalSignExtInfo = Marshal.AllocHGlobal(Marshal.SizeOf(digitalSignExtInfo));
                try
                {
                    //Marshal.StructureToPtr(digitalSignExtInfo, pDigitalSignExtInfo, false);
                    //digitalSignInfo.pSignExtInfo = pDigitalSignExtInfo;

                    IntPtr pSignContext = IntPtr.Zero;
                    if (!CryptUIWizDigitalSign(CRYPTUI_WIZ_NO_UI, IntPtr.Zero, string.Empty, ref digitalSignInfo, ref pSignContext))
                    {
                        throw new Win32Exception(Marshal.GetLastWin32Error(), "CryptUIWizDigitalSign");
                    }

                    if (!CryptUIWizFreeDigitalSignContext(pSignContext))
                        throw new Win32Exception(Marshal.GetLastWin32Error(), "CryptUIWizFreeDigitalSignContext");
                }
                finally
                {
                    Marshal.FreeHGlobal(digitalSignInfo.pSignExtInfo);
                }
            }
            catch (Win32Exception ex)
            {
                Console.WriteLine(string.Format("{0} {1:x}", ex.Message, ex.NativeErrorCode));
                Environment.Exit(1);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                Environment.Exit(1);
            }
        }
    }
}